from .._tpc import _TPC

class TPCH(_TPC):
    """
    TPC-H Benchmark Implementation
    """
    TPC_BENCHMARK_VARIANT = 'H'
    TABLE_REGISTRY = [
        'customer', 'lineitem', 'nation', 'orders', 'part',
        'partsupp', 'region', 'supplier'
    ]
    QUERY_REGISTRY = [
        'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8', 'q9', 'q10',
        'q11', 'q12', 'q13', 'q14', 'q15', 'q16', 'q17', 'q18', 'q19', 'q20',
        'q21', 'q22'
    ]
    DDL_FILE_NAME = 'ddl_v3.0.1.sql'


from abc import ABC

class BaseEngine(ABC):
    SQLGLOT_DIALECT = None
    REQUIRED_READ_ENDPOINT = None
    REQUIRED_WRITE_ENDPOINT = None
    
    def __init__(self):
        """
        Base Engine
        """
                  
    def get_total_cores(self):
        import os
        cores = os.cpu_count()
        return cores
    
    def get_compute_size(self):
        cores = self.get_total_cores()
        return f"{cores}vCore"
    
    def append_array_to_delta(self, abfss_path: str, array: list):
        """
        Append an array to a Delta table.
        """
        import pyarrow as pa
        from ..engines.delta_rs import DeltaRs
        results_table = pa.Table.from_pylist(array)
        DeltaRs().write_deltalake(
            abfss_path, 
            results_table, 
            mode="append",
            engine="rust"
        )



from ._tpc import _TPCDataGenerator
class TPCHDataGenerator(_TPCDataGenerator):
    """
    TPC-H Data Generator class.
    This class is a wrapper of the DuckDB TPC-H data generation utility.
    """
    GEN_UTIL = 'dbgen'


import posixpath
import os
import importlib.util


class _TPCDataGenerator:
    """
    TPC Data Generator class.
    This class is a wrapper of the DuckDB TPC[DS/H] data generation utility.
    """
    GEN_UTIL = ''

    def __init__(self, scale_factor: int, target_mount_folder_path: str = None, target_row_group_size_mb: int = 128):
        """
        Initialize the TPC data generator with a scale factor.

        :param scale_factor: The scale factor for the data generation.
        """
        self.scale_factor = scale_factor
        self.target_mount_folder_path = target_mount_folder_path
        self.target_row_group_size_mb = target_row_group_size_mb

        if importlib.util.find_spec("duckdb") is None:
            raise ImportError(
                "DuckDB is used for data generation but is not installed. Install using `%pip install lakebench[duckdb]` or `%pip install lakebench[datagen]`"
            )

    def run(self):
        """
        Generate TPC[DS/H] parquet data based on the input scale factor.
        """
        import duckdb
        import pyarrow.parquet as pq

        con = duckdb.connect()
        print(f"Generating in-memory tables")
        con.execute(f"CALL {self.GEN_UTIL}(sf={self.scale_factor})")
        tables = [row[0] for row in con.execute("SHOW TABLES").fetchall()]
        print(f"Generated in-memory tables: {tables}")

        for table in tables:
            sample_file = posixpath.join(self.target_mount_folder_path, f"{table}_sample.parquet")
            full_folder_path = posixpath.join(self.target_mount_folder_path, table)
            os.makedirs(self.target_mount_folder_path, exist_ok=True)
            # Write a sample for row size estimation
            print(f"\nSampling {table} to evaluate row count to target {self.target_row_group_size_mb}mb row groups...")
            con.execute(f"""
                COPY (SELECT * FROM {table} LIMIT 1000000)
                TO '{sample_file}'
                (FORMAT 'parquet')
            """)

            pf = pq.ParquetFile(sample_file)
            rg = pf.metadata.row_group(0)
            avg_row_size = rg.total_byte_size / rg.num_rows
            #print(f"{table} sample: {rg.num_rows} rows, {rg.total_byte_size / (1024*1024):.2f} MB")
            #print(f"Avg row size: {avg_row_size:.2f} bytes")
            target_size_bytes = self.target_row_group_size_mb * 1024 * 1024
            target_rows = int(target_size_bytes / avg_row_size)
            #print(f"Target ROW_GROUP_SIZE for ~{self.target_row_group_size_mb} MB: {target_rows} rows")

            # Write full table
            print(f"Writing {table} to {full_folder_path} with ROW_GROUP_SIZE {target_rows}...")
            con.execute(f"""
                COPY {table} TO '{full_folder_path}'
                (FORMAT 'parquet', ROW_GROUP_SIZE {target_rows}, PER_THREAD_OUTPUT)
            """)

            con.execute(f"DROP TABLE {table}")
            os.remove(sample_file)


from typing import Optional
from ..base import BaseBenchmark

from .engine_impl.spark import SparkELTBench
from .engine_impl.duckdb import DuckDBELTBench
from .engine_impl.daft import DaftELTBench
from .engine_impl.polars import PolarsELTBench

from ...engines.base import BaseEngine
from ...engines.spark import Spark
from ...engines.duckdb import DuckDB
from ...engines.daft import Daft
from ...engines.polars import Polars

import posixpath


class ELTBench(BaseBenchmark):
    """
    LightMode: minimal benchmark for quick comparisons.
    Includes basic ELT actions: load data, simple transforms, incremental processing, maintenance jobs, small query.
    """

    BENCHMARK_IMPL_REGISTRY = {
        Spark: SparkELTBench,
        DuckDB: DuckDBELTBench,
        Daft: DaftELTBench,
        Polars: PolarsELTBench
    }
    MODE_REGISTRY = ['light', 'full']
    TABLE_REGISTRY = [
        'call_center', 'catalog_page', 'catalog_returns', 'catalog_sales',
        'customer', 'customer_address', 'customer_demographics', 'date_dim',
        'household_demographics', 'income_band', 'inventory', 'item',
        'promotion', 'reason', 'ship_mode', 'store', 'store_returns',
        'store_sales', 'time_dim', 'warehouse', 'web_page', 'web_returns',
        'web_sales', 'web_site'
    ]

    def __init__(
            self, 
            engine: BaseEngine, 
            scenario_name: str,
            tpcds_parquet_mount_path: Optional[str] = None,
            tpcds_parquet_abfss_path: Optional[str] = None,
            result_abfss_path: Optional[str] = None,
            save_results: bool = False
            ):
        super().__init__(engine, scenario_name, result_abfss_path, save_results)
        for base_engine, benchmark_impl in self.BENCHMARK_IMPL_REGISTRY.items():
            if isinstance(engine, base_engine):
                self.benchmark_impl_class = benchmark_impl
                if self.benchmark_impl_class is None:
                    raise ValueError(
                        f"No benchmark implementation registered for engine type: {type(engine).__name__} "
                        f"in benchmark '{self.__class__.__name__}'."
                    )
                break
        else:
            raise ValueError(
                f"No benchmark implementation registered for engine type: {type(engine).__name__} "
                f"in benchmark '{self.__class__.__name__}'."
            )
        
        if isinstance(engine, Daft):
            if tpcds_parquet_mount_path is None:
                raise ValueError("parquet_mount_path must be provided for Daft engine.")
        self.source_data_path = tpcds_parquet_mount_path or tpcds_parquet_abfss_path
        self.engine = engine
        self.scenario_name = scenario_name
        self.benchmark_impl = self.benchmark_impl_class(
            self.engine
        )

        match engine.REQUIRED_READ_ENDPOINT:
            case 'mount':
                if tpcds_parquet_mount_path is None:
                    raise ValueError(f"parquet_mount_path must be provided for {type(engine).__name__} engine.")
                self.source_data_path = tpcds_parquet_mount_path
            case 'abfss':
                if tpcds_parquet_abfss_path is None:
                    raise ValueError(f"parquet_abfss_path must be provided for {type(engine).__name__} engine.")
                self.source_data_path = tpcds_parquet_abfss_path
            case _:
                if tpcds_parquet_mount_path is None and tpcds_parquet_abfss_path is None:
                    raise ValueError(
                        f"Either parquet_mount_path or parquet_abfss_path must be provided for {type(engine).__name__} engine."
                    )
                self.source_data_path = tpcds_parquet_abfss_path or tpcds_parquet_mount_path

    def run(self, mode: str = 'light'):

        match mode:
            case 'light':
                self.run_light_mode()
            case 'full':
                raise NotImplementedError("Full mode is not implemented yet.")
            case _:
                raise ValueError(f"Mode '{mode}' is not supported. Supported modes: {self.MODE_REGISTRY}.")

    def run_light_mode(self):
        for table_name in ('store_sales', 'date_dim', 'store', 'item', 'customer'):
            with self.timer(phase="Read parquet, write delta (x5)", test_item=table_name, engine=self.engine):
                self.engine.load_parquet_to_delta(
                    parquet_folder_path=posixpath.join(self.source_data_path, table_name), 
                    table_name=table_name
                )
        with self.timer(phase="Create fact table", test_item='total_sales_fact', engine=self.engine):
            self.benchmark_impl.create_total_sales_fact()

        for _ in range(3):
            with self.timer(phase="Merge 0.1% into fact table (3x)", test_item='total_sales_fact', engine=self.engine):
                self.benchmark_impl.merge_percent_into_total_sales_fact(0.001)

        with self.timer(phase="OPTIMIZE", test_item='total_sales_fact', engine=self.engine):
            self.engine.optimize_table('total_sales_fact')

        with self.timer(phase="VACUUM", test_item='total_sales_fact', engine=self.engine):
            self.engine.vacuum_table('total_sales_fact', retain_hours=0, retention_check=False)

        with self.timer(phase="Ad-hoc query (small result aggregation)", test_item='total_sales_fact', engine=self.engine):
            self.benchmark_impl.query_total_sales_fact()

        self.post_results()




from ._tpc import _TPCDataGenerator
class TPCDSDataGenerator(_TPCDataGenerator):
    """
    TPC-DS Data Generator class.
    This class is a wrapper of the DuckDB TPC-DS data generation utility.
    """
    GEN_UTIL = 'dsdgen'


from abc import ABC, abstractmethod
from typing import Dict, Type, Optional
import uuid
from datetime import datetime
from ..utils.timer import timer

class BaseBenchmark(ABC):
    BENCHMARK_IMPL_REGISTRY: Dict[Type, Type] = {}

    def __init__(self, engine, scenario_name: str, result_abfss_path: Optional[str], save_results: bool = False):
        self.engine = engine
        self.scenario_name = scenario_name
        self.result_abfss_path = result_abfss_path
        self.save_results = save_results
        self.header_detail_dict = {
            'run_id': str(uuid.uuid1()),
            'run_datetime': datetime.now().strftime('%Y-%m-%dT%H:%M:%S'),
            'engine': type(engine).__name__,
            'benchmark': self.__class__.__name__,
            'scenario': scenario_name,
            'total_cores': self.engine.get_total_cores(),
            'compute_size': self.engine.get_compute_size()
        }
        self.timer = timer
        self.results = []

    @abstractmethod
    def run(self):
        pass

    def post_results(self):
        result_array = [
            {
                **self.header_detail_dict,
                'phase': phase,
                'test_item': test_item,
                'start_datetime': start_datetime,
                "duration_sec": duration_ms / 1000,
                'duration_ms': duration_ms,
                'iteration': iteration,
                'success': success,
                'error_message': error_message
            }
            for phase, test_item, start_datetime, duration_ms, iteration, success, error_message in self.timer.results
        ]

        if self.save_results:
            if self.result_abfss_path is None:
                raise ValueError("result_abfss_path must be provided if save_results is True.")
            else:
                self.engine.append_array_to_delta(self.result_abfss_path, result_array)

        self.results.append(result_array)
        self.timer.clear_results()


from typing import Optional
from ..base import BaseBenchmark
from ...utils.query_utils import transpile_and_qualify_query

from ...engines.base import BaseEngine
from ...engines.spark import Spark
from ...engines.duckdb import DuckDB
from ...engines.daft import Daft
from ...engines.polars import Polars

import importlib.resources
import posixpath

class _TPC(BaseBenchmark):
    """
    """
    BENCHMARK_IMPL_REGISTRY = {
        Spark: None,
        DuckDB: None,
        Daft: None,
        Polars: None
    }
    MODE_REGISTRY = ['load', 'query', 'power_test']
    TPC_BENCHMARK_VARIANT = ''
    TABLE_REGISTRY = [
        'call_center', 'catalog_page', 'catalog_returns', 'catalog_sales',
        'customer', 'customer_address', 'customer_demographics', 'date_dim',
        'household_demographics', 'income_band', 'inventory', 'item',
        'promotion', 'reason', 'ship_mode', 'store', 'store_returns',
        'store_sales', 'time_dim', 'warehouse', 'web_page', 'web_returns',
        'web_sales', 'web_site'
    ]
    QUERY_REGISTRY = [
        'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8', 'q9', 'q10',
        'q11', 'q12', 'q13', 'q14a', 'q14b', 'q15', 'q16', 'q17', 'q18', 'q19', 'q20',
        'q21', 'q22', 'q23a', 'q23b', 'q24a', 'q24b', 'q25', 'q26', 'q27', 'q28', 'q29', 'q30',
        'q31', 'q32', 'q33', 'q34', 'q35', 'q36', 'q37', 'q38', 'q39a', 'q39b', 'q40',
        'q41', 'q42', 'q43', 'q44', 'q45', 'q46', 'q47', 'q48', 'q49', 'q50',
        'q51', 'q52', 'q53', 'q54', 'q55', 'q56', 'q57', 'q58', 'q59', 'q60',
        'q61', 'q62', 'q63', 'q64', 'q65', 'q66', 'q67', 'q68', 'q69', 'q70',
        'q71', 'q72', 'q73', 'q74', 'q75', 'q76', 'q77', 'q78', 'q79', 'q80',
        'q81', 'q82', 'q83', 'q84', 'q85', 'q86', 'q87', 'q88', 'q89', 'q90',
        'q91', 'q92', 'q93', 'q94', 'q95', 'q96', 'q97', 'q98', 'q99'
    ]
    DDL_FILE_NAME = ''

    def __init__(
            self, 
            engine: BaseEngine, 
            scenario_name: str,
            query_list: Optional[list[str]] | None = None,
            parquet_mount_path: Optional[str] = None,
            parquet_abfss_path: Optional[str] = None,
            result_abfss_path: Optional[str] = None,
            save_results: bool = False
            ):
        super().__init__(engine, scenario_name, result_abfss_path, save_results)
        if query_list is not None:
            expanded_query_list = []
            for query in query_list:
                if query == '*':
                    expanded_query_list.extend(self.QUERY_REGISTRY)  # Replace '*' with all queries
                else:
                    expanded_query_list.append(query)
            query_set = set(query_list)
            if not query_set.issubset(self.QUERY_REGISTRY):
                unsupported_queries = query_set - set(self.QUERY_REGISTRY)
                raise ValueError(f"Query list contains unsupported queries: {unsupported_queries}. Supported queries: {self.QUERY_REGISTRY}.")
            self.query_list = expanded_query_list
        else:
            self.query_list = self.QUERY_REGISTRY

        for base_engine, benchmark_impl in self.BENCHMARK_IMPL_REGISTRY.items():
            if isinstance(engine, base_engine):
                self.benchmark_impl_class = benchmark_impl
                break
        else:
            raise ValueError(
                f"No benchmark implementation registered for engine type: {type(engine).__name__} "
                f"in benchmark '{self.__class__.__name__}'."
            )

        self.engine = engine
        self.scenario_name = scenario_name

        match engine.REQUIRED_READ_ENDPOINT:
            case 'mount':
                if parquet_mount_path is None:
                    raise ValueError(f"parquet_mount_path must be provided for {type(engine).__name__} engine.")
                self.source_data_path = parquet_mount_path
            case 'abfss':
                if parquet_abfss_path is None:
                    raise ValueError(f"parquet_abfss_path must be provided for {type(engine).__name__} engine.")
                self.source_data_path = parquet_abfss_path
            case _:
                if parquet_mount_path is None and parquet_abfss_path is None:
                    raise ValueError(
                        f"Either parquet_mount_path or parquet_abfss_path must be provided for {type(engine).__name__} engine."
                    )
                self.source_data_path = parquet_abfss_path or parquet_mount_path

    def run(self, mode: str = 'power_test'):
        match mode:
            case 'load':
                self._run_load_test()
            case 'query':
                self._run_query_test()
            case 'power_test':
                self._run_power_test()
            case _:
                raise ValueError(f"Unknown mode '{mode}'. Supported modes: {self.MODE_REGISTRY}.")
    
    def _prepare_schema(self):
        self.engine.create_schema_if_not_exists(drop_before_create=True)
        with importlib.resources.path(f"lakebench.benchmarks.tpc{self.TPC_BENCHMARK_VARIANT.lower()}.resources.ddl", self.DDL_FILE_NAME) as ddl_path:
            with open(ddl_path, 'r') as ddl_file:
                ddl = ddl_file.read()
            
        statements = [s for s in ddl.split(';') if len(s) > 7]
        for statement in statements:
            self.engine.execute_sql_statement(statement)

    def _run_load_test(self):
        if isinstance(self.engine, Spark):
            self._prepare_schema()
        for table_name in self.TABLE_REGISTRY:
            with self.timer(phase="Load", test_item=table_name, engine=self.engine):
                self.engine.load_parquet_to_delta(
                    parquet_folder_path=posixpath.join(self.source_data_path,table_name), 
                    table_name=table_name
                )
        self.post_results()

    def _run_query_test(self):
        if isinstance(self.engine, (DuckDB, Daft, Polars)):
            for table_name in self.TABLE_REGISTRY:
                self.engine.register_table(table_name)
        for query_name in self.query_list:
            with importlib.resources.path(f"lakebench.benchmarks.tpc{self.TPC_BENCHMARK_VARIANT.lower()}.resources.queries", f'{query_name}.sql') as query_path:
                with open(query_path, 'r') as query_file:
                    query = query_file.read()

            prepped_query = transpile_and_qualify_query(
                query=query, 
                from_dialect='spark', 
                to_dialect=self.engine.SQLGLOT_DIALECT, 
                catalog=self.engine.catalog_name,
                schema=self.engine.schema_name
                )
            with self.timer(phase="Query", test_item=query_name, engine=self.engine):
                execute_query = self.engine.execute_sql_query(prepped_query)
        self.post_results()

    def _run_power_test(self):
        self._run_load_test()
        self._run_query_test()